@@ -1,7 +1,7 @@
 /******************************************************************************
 * The MIT License (MIT)
 *
-* Copyright (c) 2019-2021 Baldur Karlsson
+* Copyright (c) 2019-2022 Baldur Karlsson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
@@ -24,6 +24,24 @@
 
 #include "../test_common.h"
 
+std::string VKFullscreenQuadVertex = R"EOSHADER(
+
+#version 460 core
+
+void main()
+{
+	vec2 positions[] = {
+		vec2(-1.0f,  1.0f),
+		vec2( 1.0f,  1.0f),
+		vec2(-1.0f, -1.0f),
+		vec2( 1.0f, -1.0f),
+	};
+
+	gl_Position = vec4(positions[gl_VertexIndex], 0, 1);
+}
+
+)EOSHADER";
+
 static std::string common = R"EOSHADER(
 
 #version 460 core
@@ -84,8 +102,12 @@ void main()
 
 #if defined(WIN32)
 #include "../win32/win32_window.h"
-#else
+#elif defined(__linux__)
 #include "../linux/linux_window.h"
+#elif defined(__APPLE__)
+#include "../apple/apple_window.h"
+#else
+#error UNKNOWN PLATFORM
 #endif
 
 static VkBool32 VKAPI_PTR vulkanCallback(VkDebugUtilsMessageSeverityFlagBitsEXT messageSeverity,
@@ -139,10 +161,16 @@ void VulkanGraphicsTest::Prepare(int argc, char **argv)
 
 #if defined(WIN32)
       enabledInstExts.push_back(VK_KHR_WIN32_SURFACE_EXTENSION_NAME);
-#else
+#elif defined(__linux__)
       enabledInstExts.push_back(VK_KHR_XCB_SURFACE_EXTENSION_NAME);
 
       X11Window::Init();
+#elif defined(__APPLE__)
+      enabledInstExts.push_back(VK_MVK_MACOS_SURFACE_EXTENSION_NAME);
+
+      AppleWindow::Init();
+#else
+#error UNKNOWN PLATFORM
 #endif
 
       std::vector<const char *> optInstExts;
@@ -548,6 +576,45 @@ void VulkanGraphicsTest::Prepare(int argc, char **argv)
       }
     }
   }
+
+  std::vector<VkQueueFamilyProperties> queueProps;
+  vkh::getQueueFamilyProperties(queueProps, phys);
+
+  // if no queue has been selected, find it now
+  if(queueFamilyIndex == ~0U)
+  {
+    // try to find an exact match first
+    for(uint32_t q = 0; q < queueProps.size(); q++)
+    {
+      VkQueueFlags flags = queueProps[q].queueFlags;
+
+      if(flags == queueFlagsRequired)
+      {
+        queueFamilyIndex = q;
+        queueCount = 1;
+        break;
+      }
+    }
+  }
+
+  if(queueFamilyIndex == ~0U)
+  {
+    // if we didn't find an exact match, look for any that does satisfy what we want
+    for(uint32_t q = 0; q < queueProps.size(); q++)
+    {
+      VkQueueFlags flags = queueProps[q].queueFlags;
+
+      if(((flags & queueFlagsRequired) == queueFlagsRequired) && ((flags & queueFlagsBanned) == 0))
+      {
+        queueFamilyIndex = q;
+        queueCount = 1;
+        break;
+      }
+    }
+  }
+
+  if(queueFamilyIndex == ~0U)
+    Avail = "No satisfactory queue family available";
 }
 
 bool VulkanGraphicsTest::Init()
@@ -565,32 +632,6 @@ bool VulkanGraphicsTest::Init()
         NULL, &debugUtilsMessenger));
   }
 
-  std::vector<VkQueueFamilyProperties> queueProps;
-  vkh::getQueueFamilyProperties(queueProps, phys);
-
-  VkQueueFlags required = VK_QUEUE_GRAPHICS_BIT | VK_QUEUE_COMPUTE_BIT;
-
-  // if no queue has been selected, find it now
-  if(queueFamilyIndex == ~0U)
-  {
-    for(uint32_t q = 0; q < queueProps.size(); q++)
-    {
-      VkQueueFlags flags = queueProps[q].queueFlags;
-      if((flags & required) == required)
-      {
-        queueFamilyIndex = q;
-        queueCount = 1;
-        break;
-      }
-    }
-  }
-
-  if(queueFamilyIndex == ~0U)
-  {
-    TEST_ERROR("No graphics/compute queues available");
-    return false;
-  }
-
   std::vector<VkExtensionProperties> supportedExts;
   CHECK_VKR(vkh::enumerateDeviceExtensionProperties(supportedExts, phys, NULL));
 
@@ -620,13 +661,17 @@ bool VulkanGraphicsTest::Init()
   volkLoadDevice(device);
 
   vkGetDeviceQueue(device, queueFamilyIndex, 0, &queue);
-  mainWindow = MakeWindow(screenWidth, screenHeight, "Autotesting");
 
-  if(!mainWindow->Initialised())
+  if(!headless)
   {
-    TEST_ERROR("Error creating surface");
-    return false;
-  };
+    mainWindow = MakeWindow(screenWidth, screenHeight, "Autotesting");
+
+    if(!mainWindow->Initialised())
+    {
+      TEST_ERROR("Error creating surface");
+      return false;
+    }
+  }
 
   VmaVulkanFunctions funcs = {
       vkGetPhysicalDeviceProperties,
@@ -659,7 +704,10 @@ bool VulkanGraphicsTest::Init()
 
   vmaCreateAllocator(&allocInfo, &allocator);
 
-  TEST_LOG("Running Vulkan test on %s", physProperties.deviceName);
+  TEST_LOG("Running Vulkan test on %s (version %d.%d)", physProperties.deviceName,
+           VK_VERSION_MAJOR(physProperties.apiVersion), VK_VERSION_MINOR(physProperties.apiVersion));
+
+  headlessCmds = new VulkanCommands(this);
 
   return true;
 }
@@ -668,8 +716,12 @@ VulkanWindow *VulkanGraphicsTest::MakeWindow(int width, int height, const char *
 {
 #if defined(WIN32)
   GraphicsWindow *platWin = new Win32Window(width, height, title);
-#else
+#elif defined(__linux__)
   GraphicsWindow *platWin = new X11Window(width, height, 0, title);
+#elif defined(__APPLE__)
+  GraphicsWindow *platWin = new AppleWindow(width, height, title);
+#else
+#error UNKNOWN PLATFORM
 #endif
 
   return new VulkanWindow(this, platWin);
@@ -719,6 +771,9 @@ void VulkanGraphicsTest::Shutdown()
 
     vmaDestroyAllocator(allocator);
 
+    if(headlessCmds)
+      delete headlessCmds;
+
     delete mainWindow;
 
     vkDestroyDevice(device, NULL);
@@ -770,27 +825,17 @@ void VulkanGraphicsTest::FinishUsingBackbuffer(VkCommandBuffer cmd, VkAccessFlag
 }
 
 void VulkanGraphicsTest::Submit(int index, int totalSubmits, const std::vector<VkCommandBuffer> &cmds,
-                                const std::vector<VkCommandBuffer> &seccmds, VulkanWindow *window,
-                                VkQueue q, bool sync2)
+                                const std::vector<VkCommandBuffer> &seccmds)
 {
-  if(window == NULL)
-    window = mainWindow;
-
-  if(q == VK_NULL_HANDLE)
-    q = queue;
-
-  window->Submit(index, totalSubmits, cmds, seccmds, q, sync2);
+  if(mainWindow)
+    mainWindow->Submit(index, totalSubmits, cmds, seccmds, queue);
+  else
+    headlessCmds->Submit(cmds, seccmds, queue, VK_NULL_HANDLE, VK_NULL_HANDLE);
 }
 
-void VulkanGraphicsTest::Present(VulkanWindow *window, VkQueue q)
+void VulkanGraphicsTest::Present()
 {
-  if(!window)
-    window = mainWindow;
-
-  if(q == VK_NULL_HANDLE)
-    q = queue;
-
-  window->Present(q);
+  mainWindow->Present(queue);
 }
 
 VkPipelineShaderStageCreateInfo VulkanGraphicsTest::CompileShaderModule(
@@ -826,25 +871,10 @@ VkCommandBuffer VulkanGraphicsTest::GetCommandBuffer(VkCommandBufferLevel level,
   if(window == NULL)
     window = mainWindow;
 
-  return window->GetCommandBuffer(level);
-}
+  if(window)
+    return window->GetCommandBuffer(level);
 
-VkCommandBuffer VulkanWindow::GetCommandBuffer(VkCommandBufferLevel level)
-{
-  std::vector<VkCommandBuffer> &buflist = freeCommandBuffers[level];
-
-  if(buflist.empty())
-  {
-    buflist.resize(4);
-
-    CHECK_VKR(vkAllocateCommandBuffers(
-        m_Test->device, vkh::CommandBufferAllocateInfo(cmdPool, 4, level), &buflist[0]));
-  }
-
-  VkCommandBuffer ret = buflist.back();
-  buflist.pop_back();
-
-  return ret;
+  return headlessCmds->GetCommandBuffer(level);
 }
 
 template <>
@@ -877,6 +907,12 @@ void VulkanGraphicsTest::setName(VkBuffer obj, const std::string &name)
   setName(VK_OBJECT_TYPE_BUFFER, (uint64_t)obj, name);
 }
 
+template <>
+void VulkanGraphicsTest::setName(VkSemaphore obj, const std::string &name)
+{
+  setName(VK_OBJECT_TYPE_SEMAPHORE, (uint64_t)obj, name);
+}
+
 void VulkanGraphicsTest::setName(VkObjectType objType, uint64_t obj, const std::string &name)
 {
   if(vkSetDebugUtilsObjectNameEXT)
@@ -1013,22 +1049,36 @@ VkDescriptorSet VulkanGraphicsTest::allocateDescriptorSet(VkDescriptorSetLayout
   // failed to allocate, create a new pool and push it
   {
     VkDescriptorPool pool = VK_NULL_HANDLE;
+
+    std::vector<VkDescriptorPoolSize> poolSizes = {
+        {VK_DESCRIPTOR_TYPE_SAMPLER, 1024},
+        {VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1024},
+        {VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, 1024},
+        {VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, 1024},
+        {VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, 1024},
+        {VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, 1024},
+        {VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1024},
+        {VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1024},
+        {VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, 1024},
+        {VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, 1024},
+        {VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, 1024},
+    };
+
+    VkDescriptorPoolInlineUniformBlockCreateInfo inlineCreateInfo = {
+        VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_INLINE_UNIFORM_BLOCK_CREATE_INFO,
+    };
+    void *next = NULL;
+
+    if(hasExt(VK_EXT_INLINE_UNIFORM_BLOCK_EXTENSION_NAME) || devVersion >= VK_MAKE_VERSION(1, 3, 0))
+    {
+      poolSizes.push_back({VK_DESCRIPTOR_TYPE_INLINE_UNIFORM_BLOCK, 128 * 4096});
+
+      inlineCreateInfo.maxInlineUniformBlockBindings = 1024;
+      next = &inlineCreateInfo;
+    }
+
     CHECK_VKR(vkCreateDescriptorPool(
-        device, vkh::DescriptorPoolCreateInfo(128,
-                                              {
-                                                  {VK_DESCRIPTOR_TYPE_SAMPLER, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_UNIFORM_TEXEL_BUFFER, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, 1024},
-                                                  {VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, 1024},
-                                              }),
-        NULL, &pool));
+        device, vkh::DescriptorPoolCreateInfo(128, poolSizes).next(next), NULL, &pool));
     descPools.push_back(pool);
 
     // this must succeed or we can't continue.
@@ -1111,7 +1161,117 @@ VkSampler VulkanGraphicsTest::createSampler(const VkSamplerCreateInfo *info)
   return ret;
 }
 
+VulkanCommands::VulkanCommands(VulkanGraphicsTest *test)
+{
+  m_Test = test;
+
+  CHECK_VKR(vkCreateCommandPool(
+      m_Test->device, vkh::CommandPoolCreateInfo(VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT,
+                                                 m_Test->queueFamilyIndex),
+      NULL, &cmdPool));
+}
+
+VulkanCommands::~VulkanCommands()
+{
+  vkDestroyCommandPool(m_Test->device, cmdPool, NULL);
+
+  for(VkFence fence : fences)
+    vkDestroyFence(m_Test->device, fence, NULL);
+}
+
+VkCommandBuffer VulkanCommands::GetCommandBuffer(VkCommandBufferLevel level)
+{
+  std::vector<VkCommandBuffer> &buflist = freeCommandBuffers[level];
+
+  if(buflist.empty())
+  {
+    buflist.resize(4);
+
+    CHECK_VKR(vkAllocateCommandBuffers(
+        m_Test->device, vkh::CommandBufferAllocateInfo(cmdPool, 4, level), &buflist[0]));
+  }
+
+  VkCommandBuffer ret = buflist.back();
+  buflist.pop_back();
+
+  return ret;
+}
+
+void VulkanCommands::Submit(const std::vector<VkCommandBuffer> &cmds,
+                            const std::vector<VkCommandBuffer> &seccmds, VkQueue q,
+                            VkSemaphore wait, VkSemaphore signal)
+{
+  VkFence fence;
+  CHECK_VKR(vkCreateFence(m_Test->device, vkh::FenceCreateInfo(), NULL, &fence));
+
+  fences.insert(fence);
+
+  if(m_Test->hasExt(VK_KHR_SYNCHRONIZATION_2_EXTENSION_NAME))
+  {
+    VkSubmitInfo2KHR submit = {VK_STRUCTURE_TYPE_SUBMIT_INFO_2_KHR};
+
+    std::vector<VkCommandBufferSubmitInfoKHR> cmdSubmits;
+    for(VkCommandBuffer cmd : cmds)
+      cmdSubmits.push_back({VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO_KHR, NULL, cmd, 0});
+
+    submit.commandBufferInfoCount = (uint32_t)cmdSubmits.size();
+    submit.pCommandBufferInfos = cmdSubmits.data();
+
+    VkSemaphoreSubmitInfoKHR waitInfo = {}, signalInfo = {};
+
+    if(wait != VK_NULL_HANDLE)
+    {
+      waitInfo.sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO_KHR;
+      waitInfo.semaphore = wait;
+      waitInfo.stageMask = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT_KHR;
+
+      submit.waitSemaphoreInfoCount = 1;
+      submit.pWaitSemaphoreInfos = &waitInfo;
+    }
+
+    if(signal != VK_NULL_HANDLE)
+    {
+      signalInfo.sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO_KHR;
+      signalInfo.semaphore = signal;
+      signalInfo.stageMask = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT_KHR;
+
+      submit.signalSemaphoreInfoCount = 1;
+      submit.pSignalSemaphoreInfos = &signalInfo;
+    }
+
+    CHECK_VKR(vkQueueSubmit2KHR(q, 1, &submit, fence));
+  }
+  else
+  {
+    VkPipelineStageFlags waitStage = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT;
+
+    VkSubmitInfo submit = vkh::SubmitInfo(cmds);
+
+    if(wait != VK_NULL_HANDLE)
+    {
+      submit.waitSemaphoreCount = 1;
+      submit.pWaitDstStageMask = &waitStage;
+      submit.pWaitSemaphores = &wait;
+    }
+
+    if(signal != VK_NULL_HANDLE)
+    {
+      submit.signalSemaphoreCount = 1;
+      submit.pSignalSemaphores = &signal;
+    }
+
+    CHECK_VKR(vkQueueSubmit(q, 1, &submit, fence));
+  }
+
+  for(const VkCommandBuffer &cmd : cmds)
+    pendingCommandBuffers[0].push_back(std::make_pair(cmd, fence));
+
+  for(const VkCommandBuffer &cmd : seccmds)
+    pendingCommandBuffers[1].push_back(std::make_pair(cmd, fence));
+}
+
 VulkanWindow::VulkanWindow(VulkanGraphicsTest *test, GraphicsWindow *win)
+    : GraphicsWindow(win->title), VulkanCommands(test)
 {
   m_Test = test;
   m_Win = win;
@@ -1119,15 +1279,14 @@ VulkanWindow::VulkanWindow(VulkanGraphicsTest *test, GraphicsWindow *win)
   {
     std::lock_guard<std::mutex> lock(m_Test->mutex);
 
-    CHECK_VKR(vkCreateCommandPool(
-        m_Test->device, vkh::CommandPoolCreateInfo(VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT),
-        NULL, &cmdPool));
-
     CHECK_VKR(
         vkCreateSemaphore(m_Test->device, vkh::SemaphoreCreateInfo(), NULL, &renderStartSemaphore));
     CHECK_VKR(
         vkCreateSemaphore(m_Test->device, vkh::SemaphoreCreateInfo(), NULL, &renderEndSemaphore));
 
+    test->setName(renderStartSemaphore, title + " renderStartSemaphore");
+    test->setName(renderEndSemaphore, title + " renderEndSemaphore");
+
 #if defined(WIN32)
     VkWin32SurfaceCreateInfoKHR createInfo;
 
@@ -1138,7 +1297,7 @@ VulkanWindow::VulkanWindow(VulkanGraphicsTest *test, GraphicsWindow *win)
     createInfo.hinstance = GetModuleHandleA(NULL);
 
     vkCreateWin32SurfaceKHR(m_Test->instance, &createInfo, NULL, &surface);
-#else
+#elif defined(__linux__)
     VkXcbSurfaceCreateInfoKHR createInfo;
 
     createInfo.sType = VK_STRUCTURE_TYPE_XCB_SURFACE_CREATE_INFO_KHR;
@@ -1148,6 +1307,16 @@ VulkanWindow::VulkanWindow(VulkanGraphicsTest *test, GraphicsWindow *win)
     createInfo.window = ((X11Window *)win)->xcb.window;
 
     vkCreateXcbSurfaceKHR(m_Test->instance, &createInfo, NULL, &surface);
+#elif defined(__APPLE__)
+    VkMacOSSurfaceCreateInfoMVK createInfo;
+
+    createInfo.sType = VK_STRUCTURE_TYPE_MACOS_SURFACE_CREATE_INFO_MVK;
+    createInfo.pNext = NULL;
+    createInfo.flags = 0;
+    createInfo.pView = ((AppleWindow *)win)->view;
+    vkCreateMacOSSurfaceMVK(m_Test->instance, &createInfo, NULL, &surface);
+#else
+#error UNKNOWN PLATFORM
 #endif
   }
 
@@ -1164,11 +1333,6 @@ VulkanWindow::~VulkanWindow()
     vkDestroySemaphore(m_Test->device, renderStartSemaphore, NULL);
     vkDestroySemaphore(m_Test->device, renderEndSemaphore, NULL);
 
-    vkDestroyCommandPool(m_Test->device, cmdPool, NULL);
-
-    for(VkFence fence : fences)
-      vkDestroyFence(m_Test->device, fence, NULL);
-
     if(surface)
       vkDestroySurfaceKHR(m_Test->instance, surface, NULL);
   }
@@ -1299,75 +1463,57 @@ void VulkanWindow::Acquire()
 }
 
 void VulkanWindow::Submit(int index, int totalSubmits, const std::vector<VkCommandBuffer> &cmds,
-                          const std::vector<VkCommandBuffer> &seccmds, VkQueue q, bool sync2)
+                          const std::vector<VkCommandBuffer> &seccmds, VkQueue q)
 {
-  VkFence fence;
-  CHECK_VKR(vkCreateFence(m_Test->device, vkh::FenceCreateInfo(), NULL, &fence));
+  VkSemaphore signal = VK_NULL_HANDLE, wait = VK_NULL_HANDLE;
 
-  fences.insert(fence);
+  if(index == 0)
+    wait = renderStartSemaphore;
+  if(index == totalSubmits - 1)
+    signal = renderEndSemaphore;
 
-  if(sync2)
+  VulkanCommands::Submit(cmds, seccmds, q, wait, signal);
+}
+
+void VulkanWindow::MultiPresent(VkQueue queue, std::vector<VulkanWindow *> windows)
+{
+  std::vector<VkSwapchainKHR> swaps;
+  std::vector<uint32_t> idxs;
+  std::vector<VkSemaphore> waitSems;
+  std::vector<VkResult> vkrs;
+
+  for(auto it : windows)
   {
-    VkSubmitInfo2KHR submit = {VK_STRUCTURE_TYPE_SUBMIT_INFO_2_KHR};
+    if(it->swap == VK_NULL_HANDLE)
+      continue;
 
-    std::vector<VkCommandBufferSubmitInfoKHR> cmdSubmits;
-    for(VkCommandBuffer cmd : cmds)
-      cmdSubmits.push_back({VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO_KHR, NULL, cmd, 0});
-
-    submit.commandBufferInfoCount = (uint32_t)cmdSubmits.size();
-    submit.pCommandBufferInfos = cmdSubmits.data();
-
-    VkSemaphoreSubmitInfoKHR renderStart = {}, renderEnd = {};
-
-    if(index == 0)
-    {
-      renderStart.sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO_KHR;
-      renderStart.semaphore = renderStartSemaphore;
-      renderStart.stageMask = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT_KHR;
-
-      submit.waitSemaphoreInfoCount = 1;
-      submit.pWaitSemaphoreInfos = &renderStart;
-    }
-
-    if(index == totalSubmits - 1)
-    {
-      renderEnd.sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO_KHR;
-      renderEnd.semaphore = renderEndSemaphore;
-      renderEnd.stageMask = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT_KHR;
-
-      submit.signalSemaphoreInfoCount = 1;
-      submit.pSignalSemaphoreInfos = &renderEnd;
-    }
-
-    CHECK_VKR(vkQueueSubmit2KHR(q, 1, &submit, fence));
-  }
-  else
-  {
-    VkPipelineStageFlags waitStage = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT;
-
-    VkSubmitInfo submit = vkh::SubmitInfo(cmds);
-
-    if(index == 0)
-    {
-      submit.waitSemaphoreCount = 1;
-      submit.pWaitDstStageMask = &waitStage;
-      submit.pWaitSemaphores = &renderStartSemaphore;
-    }
-
-    if(index == totalSubmits - 1)
-    {
-      submit.signalSemaphoreCount = 1;
-      submit.pSignalSemaphores = &renderEndSemaphore;
-    }
-
-    CHECK_VKR(vkQueueSubmit(q, 1, &submit, fence));
+    swaps.push_back(it->swap);
+    idxs.push_back(it->imgIndex);
+    waitSems.push_back(it->renderEndSemaphore);
+    vkrs.push_back(VK_SUCCESS);
   }
 
-  for(const VkCommandBuffer &cmd : cmds)
-    pendingCommandBuffers[0].push_back(std::make_pair(cmd, fence));
+  if(swaps.empty())
+    return;
 
-  for(const VkCommandBuffer &cmd : seccmds)
-    pendingCommandBuffers[1].push_back(std::make_pair(cmd, fence));
+  VkPresentInfoKHR info = {VK_STRUCTURE_TYPE_PRESENT_INFO_KHR};
+  info.swapchainCount = (uint32_t)swaps.size();
+  info.waitSemaphoreCount = (uint32_t)waitSems.size();
+  info.pSwapchains = swaps.data();
+  info.pImageIndices = idxs.data();
+  info.pWaitSemaphores = waitSems.data();
+  info.pResults = vkrs.data();
+
+  vkQueuePresentKHR(queue, &info);
+
+  size_t i = 0;
+  for(auto it : windows)
+  {
+    if(it->swap == VK_NULL_HANDLE)
+      continue;
+
+    it->PostPresent(vkrs[i++]);
+  }
 }
 
 void VulkanWindow::Present(VkQueue queue)
@@ -1377,6 +1523,11 @@ void VulkanWindow::Present(VkQueue queue)
 
   VkResult vkr = vkQueuePresentKHR(queue, vkh::PresentInfoKHR(swap, imgIndex, &renderEndSemaphore));
 
+  PostPresent(vkr);
+}
+
+void VulkanWindow::PostPresent(VkResult vkr)
+{
   if(vkr == VK_SUBOPTIMAL_KHR || vkr == VK_ERROR_OUT_OF_DATE_KHR)
   {
     DestroySwapchain();
@@ -1388,6 +1539,13 @@ void VulkanWindow::Present(VkQueue queue)
     CHECK_VKR(queuePresentError);
   }
 
+  VulkanCommands::ProcessCompletions();
+
+  Acquire();
+}
+
+void VulkanCommands::ProcessCompletions()
+{
   std::set<VkFence> doneFences;
   std::map<VkFence, VkResult> fenceStatus;
 
@@ -1419,8 +1577,6 @@ void VulkanWindow::Present(VkQueue queue)
     vkDestroyFence(m_Test->device, *it, NULL);
     fences.erase(*it);
   }
-
-  Acquire();
 }
 
 void VulkanWindow::DestroySwapchain()
